import json
import logging
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, Iterator, List, Optional
from uuid import UUID, uuid4

from pydantic import BaseModel, Field

from cado.core.cell import Cell
from cado.core.cell_status import CellStatus
from cado.core.language import Language

logger = logging.getLogger(__name__)

CURRENT_VERSION = "0.1"


class Notebook(BaseModel):
    id: UUID = Field(default_factory=uuid4)
    name: str
    version: str = CURRENT_VERSION
    created: datetime = Field(default_factory=datetime.now)
    updated: datetime = Field(default_factory=datetime.now)
    cells: List[Cell] = []

    def update_cell_output_name(self, cell_id: UUID, output_name: str) -> None:
        """Set a notebook cell's output name.

        Args:
            cell_id (UUID): ID of the cell.
            output_name (str): Name for the output variable.
        """
        self.clear_cell(cell_id)

        output_names = set()
        for cell in self.cells:
            if cell.id != cell_id and cell.output_name != "":
                output_names.add(cell.output_name)

        cell = self.get_cell(cell_id)

        for child in self._get_children(cell):
            child.clear()

        if output_name in output_names:
            error = ValueError(f"Cell with output name \"{output_name}\" already exists in the notebook")
            self.error_cell(cell.id, error)
            cell.output_name = ""
            raise error

        cell.output_name = output_name

    def update_cell_input_names(self, cell_id: UUID, input_names: List[str]) -> None:
        """Set a notebook cell's input names.

        Args:
            cell_id (UUID): ID of the cell.
            input_names (List[str]): Names for the input variables.
        """
        self._check_no_self_ancestor(cell_id, cell_id, input_names)
        self.clear_cell(cell_id)

        output_names = set()
        for cell in self.cells:
            if cell.id != cell_id and cell.output_name != "":
                output_names.add(cell.output_name)

        cell = self.get_cell(cell_id)
        for input_name in input_names:
            if input_name == cell.output_name:
                error = ValueError("A cell cannot have its own output as input")
                self.error_cell(cell.id, error)
                cell.input_names = []
                raise error

            if input_name not in output_names:
                error = ValueError(f"No cell with output name \"{input_name}\"")
                self.error_cell(cell.id, error)
                cell.input_names = []
                raise error

        cell.input_names = input_names

    def set_cell_code(self, cell_id: UUID, code: str) -> None:
        """Set a notebook cell's code.

        Args:
            cell_id (UUID): ID of the cell.
            code (str): String of code to set on the cell.
        """
        cell = self.get_cell(cell_id)
        cell.code = code
        self.clear_cell(cell.id)

    def get_cell(self, cell_id: UUID) -> Cell:
        """Get a cell from the notebook.

        Args:
            cell_id (UUID): ID of the cell to get from the notebook.
        """
        for cell in self.cells:
            if cell.id == cell_id:
                return cell
        raise ValueError(f"No cell with ID {cell_id} found in notebook")

    def delete_cell(self, cell_id: UUID) -> None:
        """Delete a cell from the notebook.

        Args:
            cell_id (UUID): ID of the cell to delete from the notebook.
        """
        self.cells = [c for c in self.cells if c.id != cell_id]

    def add_cell(self, index: Optional[int] = None) -> UUID:
        """Add a cell to the notebook.

        Args:
            index (Optional[int]): Index where the cell should be inserted.

        Returns:
            UUID: ID of the new cell.
        """
        new_cell = Cell(output_name="")
        if index is None:
            self.cells.append(new_cell)
        else:
            self.cells.insert(index, new_cell)
        return new_cell.id

    def _get_children(self, cell: Cell) -> Iterator[Cell]:
        for other in self.cells:
            for input_name in other.input_names:
                if input_name == cell.output_name:
                    yield other

    def _get_parents(self, cell: Cell) -> Iterator[Cell]:
        for other in self.cells:
            for input_name in cell.input_names:
                if other.output_name == input_name:
                    yield other

    def _check_no_self_ancestor(
        self,
        target_cell_id: UUID,
        current_cell_id: UUID,
        input_names: List[str],
    ) -> None:
        for cell in self.cells:
            if cell.id != current_cell_id:
                for input_name in input_names:
                    if input_name == cell.output_name:
                        if cell.id == target_cell_id:
                            error = ValueError("Cycle found in cell dependencies")
                            self.error_cell(target_cell_id, error)
                            raise error
                        self._check_no_self_ancestor(target_cell_id, cell.id, cell.input_names)

    def run_cell(self, cell_id: UUID) -> None:
        """Run a cell in the notebook.

        Args:
            cell_id (UUID): ID of the cell to run.
        """
        cell = self.get_cell(cell_id)

        context: Dict[str, Any] = {}
        for parent in self._get_parents(cell):
            if parent.status in [CellStatus.EXPIRED, CellStatus.ERROR]:
                self.run_cell(parent.id)
            if parent.status == CellStatus.ERROR:
                raise ValueError("Parent cell does not have OK status, could not run child cell")
            context[parent.output_name] = parent.output

        logger.debug("Running cell %s", cell.id)
        cell.run(context)

        if cell.status == CellStatus.OK:
            for child in self._get_children(cell):
                self.run_cell(child.id)

    def update_cell_language(self, cell_id: UUID, language: Language) -> None:
        """Run a cell in the notebook.

        Args:
            cell_id (UUID): ID of the cell to run.
            language (Language): New language to use.
        """
        cell = self.get_cell(cell_id)
        cell.language = language
        cell.clear()

    def error_cell(self, cell_id: UUID, error: ValueError) -> None:
        """Error out a cell in the notebook.

        Args:
            cell_id (UUID): ID of the cell to error.
        """
        cell = self.get_cell(cell_id)
        cell.set_error(error)

        for child in self._get_children(cell):
            self.clear_cell(child.id)

    def clear_cell(self, cell_id: UUID) -> None:
        """Clear a cell in the notebook.

        Args:
            cell_id (UUID): ID of the cell to clear.
        """
        cell = self.get_cell(cell_id)
        cell.clear()

        for child in self._get_children(cell):
            self.clear_cell(child.id)

    def reorder_cells(self, cell_ids: List[UUID]) -> None:
        """Reorder cells in the notebook.

        Args:
            cell_ids (List[UUID]): List of cell IDs in the correct order.
        """
        new_cells = []
        for cell_id in cell_ids:
            cell = self.get_cell(cell_id)
            new_cells.append(cell)
        self.cells = new_cells

    def set_updated_time(self) -> None:
        """Set the updated time to now."""
        self.updated = datetime.now()

    @classmethod
    def from_filepath(cls, filepath: Path) -> "Notebook":
        """Load a notebook from a .cado notebook file.

        Args:
            filepath (Path): Filepath to a .cado notebook file.

        Returns:
            Notebook: Loaded cado notebook.
        """
        with filepath.open() as f:
            notebook_json = json.load(f)
            version = notebook_json["version"]
            if version != CURRENT_VERSION:
                raise ValueError(f"Cannot load notebook with version {version}")
            return cls.parse_obj(notebook_json)

    def to_filepath(self, filepath: Path) -> None:
        """Save a notebook to a .cado notebook file.

        Args:
            filepath (Path): Filepath to a .cado notebook file.
        """
        with filepath.open("w") as f:
            notebook_json = self.json()
            f.write(notebook_json)



import io
import json
import traceback
from contextlib import redirect_stdout, redirect_stderr
from typing import Any, Dict, List, Mapping, Optional
from uuid import UUID, uuid4

from pydantic import BaseModel, Field

from cado.core.cell_status import CellStatus
from cado.core.language import Language


class Cell(BaseModel):
    id: UUID = Field(default_factory=uuid4)
    code: str = ""

    output_name: str
    output: Optional[Any] = None
    input_names: List[str] = []
    language: Language = Language.PYTHON

    stdout: Optional[str] = None
    stderr: Optional[str] = None
    status: CellStatus = CellStatus.EXPIRED

    def run(self, context: Dict[str, Any]) -> None:
        """Run the cell.

        Returns:
            Any: The output of running the cell.
        """
        if self.code == "":
            error = ValueError(f"Code is empty for cell \"{self.id}\"")
            self.set_error(error)
            raise error

        exec_locals: Mapping[str, object] = {}
        try:

            stdout = io.StringIO()
            stderr = io.StringIO()
            with redirect_stdout(stdout), redirect_stderr(stderr):
                # pylint: disable=exec-used
                exec(self.code, context, exec_locals)

            self.stdout = stdout.getvalue().rstrip()
            self.stderr = stderr.getvalue().rstrip()
        except Exception as exc:
            error = ValueError(f"Failed to exec: {traceback.format_exc()}")
            self.set_error(error)
            raise error from exc

        if self.output_name != "":
            # Check that a variable with the cell output name was emitted by exec
            if self.output_name not in exec_locals:
                error = ValueError(
                    f"Cell name \"{self.output_name}\" was not found in exec locals for cell ({self.id})")
                self.set_error(error)
                raise error
            self.output = self.parse_output(exec_locals[self.output_name])
        self.status = CellStatus.OK

    def parse_output(self, output: Any) -> Any:
        """Parse the cells output as JSON.

        Args:
            output (Any): The cell output.

        Returns:
            Any: The cell output if it could be JSON serialized.
        """
        try:
            json.dumps(output)
        except TypeError as exc:
            error = ValueError(f"Cell outputs must be JSON serializable. {exc}.")
            self.set_error(error)
            raise error from exc
        return output

    def clear(self) -> None:
        """Clear the cell outputs and set the status to expired."""
        self.output = None
        self.stdout = None
        self.stderr = None
        self.status = CellStatus.EXPIRED

    def set_error(self, error: ValueError) -> None:
        """Clear the cell outputs and set the status to error."""
        self.output = None
        self.status = CellStatus.ERROR
        self.stderr = str(error).rstrip()



import logging
import traceback

from fastapi import APIRouter, WebSocket, WebSocketDisconnect
from fastapi.responses import JSONResponse

from cado.app.disk import save_notebook
from cado.app.message import ErrorResponse, MessageType
from cado.app.response import process_message
from cado.app.session_state import SessionState

logger = logging.getLogger(__name__)

router = APIRouter()


# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
@router.websocket(path="/stream")
async def stream_api(socket: WebSocket) -> None:
    """Websocket endpoint for streaming commands."""
    logger.info("Starting connection...")

    await socket.accept()
    logger.info("Connection open")

    session_state = SessionState()

    try:
        while True:
            message_json = await socket.receive_json()
            try:
                message_type = MessageType.from_str(message_json["type"])
                logger.debug("Received client message: %s", message_json)
                response = process_message(message_type, message_json, session_state)
                response_json = response.json()
                logger.debug("Sending server message: %s", response_json)
                await socket.send_json(response_json)
            # pylint: disable=broad-exception-caught
            except Exception as exc:
                logger.error("Exception raised during cado session loop")
                logger.error("Traceback: %s", traceback.format_exc())
                response = ErrorResponse(error=str(exc))
                response_json = response.json()
                logger.debug("Sending server message: %s", response_json)
                await socket.send_json(response_json)
    except WebSocketDisconnect:
        logger.info("Websocket disconnected")
        save_notebook(session_state)


@router.get(path="/status")
def get_status() -> JSONResponse:
    """Endpoint for app status."""
    logger.info("GET app status")
    return JSONResponse({
        'status': 'healthy',
    })



from enum import Enum
from pathlib import Path
from typing import List, Optional
from uuid import UUID

from pydantic import BaseModel

from cado.core.cell import Cell
from cado.core.language import Language
from cado.core.notebook import Notebook
from cado.core.notebook_details import NotebookDetails


class MessageType(Enum):
    # publish
    GET_NOTEBOOK = "get-notebook"
    UPDATE_CELL_CODE = "update-cell-code"
    UPDATE_CELL_OUTPUT_NAME = "update-cell-output-name"
    UPDATE_CELL_INPUT_NAMES = "update-cell-input-names"
    UPDATE_CELL_LANGUAGE = "update-cell-language"
    RUN_CELL = "run-cell"
    CLEAR_CELL = "clear-cell"
    NEW_CELL = "new-cell"
    DELETE_CELL = "delete-cell"
    REORDER_CELLS = "reorder-cells"
    LIST_NOTEBOOKS = "list-notebooks"
    NEW_NOTEBOOK = "new-notebook"
    DELETE_NOTEBOOK = "delete-notebook"
    OPEN_NOTEBOOK = "open-notebook"
    EXIT_NOTEBOOK = "exit-notebook"
    UPDATE_NOTEBOOK_NAME = "update-notebook-name"

    # subscribe
    GET_NOTEBOOK_RESPONSE = "get-notebook-response"
    GET_CELL_RESPONSE = "get-cell-response"
    ERROR_RESPONSE = "error-response"
    LIST_NOTEBOOKS_RESPONSE = "list-notebooks-response"

    @classmethod
    def from_str(cls, message_name: str) -> 'MessageType':
        """Parse MessageType from string.

        Args:
            message_name (str): Name of message type to parse.

        Raises:
            ValueError: If the name was not a valid message type.

        Returns:
            MessageType: The parsed MessageType.
        """
        for _, enum_value in cls.__members__.items():
            if enum_value.value == message_name:
                return enum_value
        raise ValueError(f"Unknown message type '{message_name}'")


class Message(BaseModel):
    type: MessageType


# region: publish


class GetNotebook(Message):
    type: MessageType = MessageType.GET_NOTEBOOK


class UpdateCellCode(Message):
    cell_id: UUID
    code: str
    type: MessageType = MessageType.UPDATE_CELL_CODE


class UpdateCellOutputName(Message):
    cell_id: UUID
    output_name: str
    type: MessageType = MessageType.UPDATE_CELL_OUTPUT_NAME


class UpdateCellInputNames(Message):
    cell_id: UUID
    input_names: List[str]
    type: MessageType = MessageType.UPDATE_CELL_INPUT_NAMES


class UpdateCellLanguage(Message):
    cell_id: UUID
    language: Language
    type: MessageType = MessageType.UPDATE_CELL_LANGUAGE


class RunCell(Message):
    cell_id: UUID
    type: MessageType = MessageType.RUN_CELL


class ClearCell(Message):
    cell_id: UUID
    type: MessageType = MessageType.CLEAR_CELL


class NewCell(Message):
    index: Optional[int]
    type: MessageType = MessageType.NEW_CELL


class DeleteCell(Message):
    cell_id: UUID
    type: MessageType = MessageType.DELETE_CELL


class ReorderCells(Message):
    cell_ids: List[UUID]
    type: MessageType = MessageType.REORDER_CELLS


class ListNotebooks(Message):
    type: MessageType = MessageType.LIST_NOTEBOOKS


class NewNotebook(Message):
    type: MessageType = MessageType.NEW_NOTEBOOK


class DeleteNotebook(Message):
    filepath: Path
    type: MessageType = MessageType.DELETE_NOTEBOOK


class OpenNotebook(Message):
    filepath: Path
    type: MessageType = MessageType.OPEN_NOTEBOOK


class ExitNotebook(Message):
    type: MessageType = MessageType.EXIT_NOTEBOOK


class UpdateNotebookName(Message):
    name: str
    type: MessageType = MessageType.UPDATE_NOTEBOOK_NAME


# endregion: publish

# region: subscribe


class GetNotebookResponse(Message):
    notebook: Optional[Notebook]
    type: MessageType = MessageType.GET_NOTEBOOK_RESPONSE


class GetCellResponse(Message):
    cell: Cell
    type: MessageType = MessageType.GET_CELL_RESPONSE


class ErrorResponse(Message):
    error: str
    type: MessageType = MessageType.ERROR_RESPONSE


class ListNotebooksResponse(Message):
    notebook_details: List[NotebookDetails]
    type: MessageType = MessageType.LIST_NOTEBOOKS_RESPONSE


# endregion: subscribe



import logging
from typing import Any
from cado.app.disk import (create_notebook, delete_existing_notebook, list_local_notebooks, rename_notebook,
                           save_notebook)

from cado.app.message import (ClearCell, DeleteCell, DeleteNotebook, ErrorResponse, ExitNotebook, GetNotebook,
                              GetNotebookResponse, ListNotebooks, ListNotebooksResponse, Message, MessageType, NewCell,
                              NewNotebook, OpenNotebook, ReorderCells, RunCell, UpdateCellCode, UpdateCellInputNames,
                              UpdateCellLanguage, UpdateCellOutputName, UpdateNotebookName)
from cado.app.session_state import SessionState
from cado.core.notebook import Notebook

logger = logging.getLogger(__name__)


# pylint: disable=too-many-locals
# pylint: disable=too-many-return-statements
# pylint: disable=too-many-branches
# pylint: disable=too-many-statements
def process_message(
    message_type: MessageType,
    message_json: Any,
    session_state: SessionState,
) -> Message:
    """Process a request message from the client, update the session state, and return a response message.

    Args:
        message_type (MessageType): Type of request message.
        message_json (Any): Raw request message object.
        session_state (SessionState): Current session state.

    Returns:
        Message: A response message.
    """
    notebook = session_state.notebook

    if message_type == MessageType.GET_NOTEBOOK:
        GetNotebook.parse_obj(message_json)
    elif message_type == MessageType.UPDATE_CELL_CODE:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        update_cell_code = UpdateCellCode.parse_obj(message_json)
        notebook.set_cell_code(update_cell_code.cell_id, update_cell_code.code)
    elif message_type == MessageType.UPDATE_CELL_OUTPUT_NAME:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        update_cell_output_name = UpdateCellOutputName.parse_obj(message_json)
        notebook.update_cell_output_name(update_cell_output_name.cell_id, update_cell_output_name.output_name)
    elif message_type == MessageType.RUN_CELL:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        run_cell = RunCell.parse_obj(message_json)
        notebook.run_cell(run_cell.cell_id)
    elif message_type == MessageType.CLEAR_CELL:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        clear_cell = ClearCell.parse_obj(message_json)
        notebook.clear_cell(clear_cell.cell_id)
    elif message_type == MessageType.NEW_CELL:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        new_cell = NewCell.parse_obj(message_json)
        notebook.add_cell(index=new_cell.index)
    elif message_type == MessageType.DELETE_CELL:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        delete_cell = DeleteCell.parse_obj(message_json)
        notebook.delete_cell(delete_cell.cell_id)
    elif message_type == MessageType.UPDATE_CELL_INPUT_NAMES:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        update_cell_input_names = UpdateCellInputNames.parse_obj(message_json)
        notebook.update_cell_input_names(update_cell_input_names.cell_id, update_cell_input_names.input_names)
    elif message_type == MessageType.UPDATE_CELL_LANGUAGE:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        update_cell_language = UpdateCellLanguage.parse_obj(message_json)
        notebook.update_cell_language(update_cell_language.cell_id, update_cell_language.language)
    elif message_type == MessageType.REORDER_CELLS:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        reorder_cells = ReorderCells.parse_obj(message_json)
        notebook.reorder_cells(reorder_cells.cell_ids)
    elif message_type == MessageType.LIST_NOTEBOOKS:
        ListNotebooks.parse_obj(message_json)
        details = list_local_notebooks()
        return ListNotebooksResponse(notebook_details=details)
    elif message_type == MessageType.OPEN_NOTEBOOK:
        open_notebook = OpenNotebook.parse_obj(message_json)
        session_state.notebook = Notebook.from_filepath(open_notebook.filepath)
        session_state.filepath = open_notebook.filepath
    elif message_type == MessageType.EXIT_NOTEBOOK:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        ExitNotebook.parse_obj(message_json)
        session_state.notebook = None
        session_state.filepath = None
    elif message_type == MessageType.UPDATE_NOTEBOOK_NAME:
        if notebook is None:
            return ErrorResponse(error=f"Can't process {message_type}, no active notebook")
        update_notebook_name = UpdateNotebookName.parse_obj(message_json)
        rename_notebook(session_state, notebook, update_notebook_name.name)
    elif message_type == MessageType.NEW_NOTEBOOK:
        NewNotebook.parse_obj(message_json)
        create_notebook(session_state)
    elif message_type == MessageType.DELETE_NOTEBOOK:
        delete_notebook = DeleteNotebook.parse_obj(message_json)
        delete_existing_notebook(delete_notebook.filepath)
        details = list_local_notebooks()
        return ListNotebooksResponse(notebook_details=details)
    else:
        logger.error("Request type did not match any known message types")
        return ErrorResponse(error=f"Unknown message type from client: {message_type}")

    save_notebook(session_state)

    return GetNotebookResponse(notebook=session_state.notebook)



from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from cado.core.notebook import Notebook


@dataclass
class SessionState:
    notebook: Optional[Notebook] = None
    filepath: Optional[Path] = None



import logging
from pathlib import Path
from typing import List

from cado.app.example import load_example_notebook
from cado.app.session_state import SessionState
from cado.core.notebook import Notebook
from cado.core.notebook_details import NotebookDetails

logger = logging.getLogger(__name__)


def save_notebook(session_state: SessionState) -> None:
    """Save the current session notebook to the disk.

    Args:
        session_state (SessionState): Current session state.
    """
    notebook = session_state.notebook
    if notebook is None:
        logger.debug("Not saving notebook to file, notebook was None")
        return

    filepath = session_state.filepath
    if filepath is None:
        logger.debug("Not saving notebook to file, filepath was None")
        return

    notebook.set_updated_time()
    logger.debug("Saving notebook to file: %s", filepath)
    notebook.to_filepath(filepath)


def create_notebook(session_state: SessionState) -> None:
    """Create a new notebook and set it as the current session notebook.
    Also save the new empty notebook to disk.

    Args:
        session_state (SessionState): Current session state.
    """
    cwd = Path.cwd()
    name = get_unique_notebook_name("notebook", cwd)
    filename = f"{name}.cado"
    filepath = cwd / filename
    notebook = Notebook(name=name)
    notebook.add_cell()
    notebook.to_filepath(filepath)
    session_state.filepath = filepath
    session_state.notebook = notebook


def get_unique_notebook_name(name: str, dirpath: Path) -> str:
    """Attempts to find a unique notebook name and avoids duplicates.

    Args:
        name (str): The desired name for the notebook.
        dirpath (Path): The directory where the notebook will be saved.

    Returns:
        str: The unique name the notebook should be saved as.
    """
    for i in range(1, 100):
        name = name if i == 1 else f"{name}-{i}"
        filename = f"{name}.cado"
        filepath = dirpath / filename
        if not filepath.exists():
            return name
    raise ValueError("Could not create a new notebook, found too many existing new notebooks")


def delete_existing_notebook(filepath: Path) -> None:
    """Delete a notebook from the disk.

    Args:
        filepath (Path): Path on disk to the notebook.
    """
    filepath.unlink()


def rename_notebook(session_state: SessionState, notebook: Notebook, name: str) -> None:
    """Rename the notebook given a new name.

    Args:
        session_state (SessionState): Current session state.
        notebook (Notebook): Notebook to update.
        name (str): Name to use when renaming the notebook.
    """
    notebook.name = name

    if name == "":
        return

    old_filepath = session_state.filepath
    if old_filepath is None:
        return

    dirpath = old_filepath.parent
    new_name = get_unique_notebook_name(name, dirpath)
    new_filepath = dirpath / f"{new_name}.cado"

    session_state.filepath = new_filepath
    save_notebook(session_state)
    delete_existing_notebook(old_filepath)


def list_local_notebooks() -> List[NotebookDetails]:
    """List all of the notebooks on the local disk.

    Returns:
        List[NotebookDetails]: A list of NotebookDetail objects.
    """
    details = []
    cwd = Path.cwd()
    for path in cwd.iterdir():
        if path.is_file():
            if str(path).endswith(".cado"):
                details.append(NotebookDetails.from_filepath(path))

    if len(details) == 0:
        filename = "example.cado"
        filepath = cwd / filename
        notebook = load_example_notebook(filename)
        notebook.to_filepath(filepath)
        details.append(NotebookDetails.from_filepath(filepath))

    return details



import logging
import pkg_resources
import webbrowser
from pathlib import Path

import click
import uvicorn

from cado.app.app import app as cado_app

logger = logging.getLogger(__name__)


@click.group()
def main() -> None:
    """Main CLI entrypoint."""


@main.command(name="up")
@click.option("--host", type=str, default="localhost")
@click.option("--port", "-p", type=int, default=8000)
@click.option("--debug", is_flag=True, default=False)
@click.option("--log-level", type=int, default=logging.INFO)
@click.option("--open/--no-open", is_flag=True, default=True)
def up_command(host: str, port: int, debug: bool, log_level: int, open: bool) -> None:
    """Command to start up cado app."""
    current_dirpath = Path(__file__).parent
    cado_string = """
                 _
                ( )
   ___   _ _   _| |  _
 / ___)/ _  )/ _  |/ _ \\
( (___( (_| | (_| | (_) )
 \\____)\\__ _)\\__ _)\\___/
"""

    print("\033[0;32m")
    print(cado_string)

    package_version = pkg_resources.get_distribution("cado").version
    print(f"\nRunning cado version {package_version}\n")

    url = f"http://{host}:{port}"
    print(f"\nOpen \033[00m [[ \033[0;34m {url} \033[00m ]] \033[0;32m in a browser to get started")

    print("\033[00m")

    if open:
        webbrowser.open(url)

    logger.info(f"Starting app from {current_dirpath}")

    package_dirpath = current_dirpath.parent.parent
    log_config_filepath = package_dirpath / "logging.yaml"
    uvicorn.run(
        cado_app,
        host=host,
        port=port,
        debug=debug,
        log_level=log_level,
        log_config=str(log_config_filepath),
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cado', 'cado.app', 'cado.cli', 'cado.core', 'cado.examples']

package_data = \
{'': ['*'],
 'cado': ['ui/.gitignore',
          'ui/.gitignore',
          'ui/.gitignore',
          'ui/.gitignore',
          'ui/.gitignore',
          'ui/.gitignore',
          'ui/.gitignore',
          'ui/.gitignore',
          'ui/.gitignore',
          'ui/.gitignore',
          'ui/dist/*',
          'ui/dist/assets/*',
          'ui/index.html',
          'ui/index.html',
          'ui/index.html',
          'ui/index.html',
          'ui/index.html',
          'ui/index.html',
          'ui/index.html',
          'ui/index.html',
          'ui/index.html',
          'ui/index.html',
          'ui/package-lock.json',
          'ui/package-lock.json',
          'ui/package-lock.json',
          'ui/package-lock.json',
          'ui/package-lock.json',
          'ui/package-lock.json',
          'ui/package-lock.json',
          'ui/package-lock.json',
          'ui/package-lock.json',
          'ui/package-lock.json',
          'ui/package.json',
          'ui/package.json',
          'ui/package.json',
          'ui/package.json',
          'ui/package.json',
          'ui/package.json',
          'ui/package.json',
          'ui/package.json',
          'ui/package.json',
          'ui/package.json',
          'ui/postcss.config.cjs',
          'ui/postcss.config.cjs',
          'ui/postcss.config.cjs',
          'ui/postcss.config.cjs',
          'ui/postcss.config.cjs',
          'ui/postcss.config.cjs',
          'ui/postcss.config.cjs',
          'ui/postcss.config.cjs',
          'ui/postcss.config.cjs',
          'ui/postcss.config.cjs',
          'ui/public/*',
          'ui/src/*',
          'ui/src/components/*',
          'ui/src/hooks/*',
          'ui/src/images/*',
          'ui/src/lib/*',
          'ui/src/lib/models/*',
          'ui/src/widgets/*',
          'ui/tailwind.config.cjs',
          'ui/tailwind.config.cjs',
          'ui/tailwind.config.cjs',
          'ui/tailwind.config.cjs',
          'ui/tailwind.config.cjs',
          'ui/tailwind.config.cjs',
          'ui/tailwind.config.cjs',
          'ui/tailwind.config.cjs',
          'ui/tailwind.config.cjs',
          'ui/tailwind.config.cjs',
          'ui/tsconfig.json',
          'ui/tsconfig.json',
          'ui/tsconfig.json',
          'ui/tsconfig.json',
          'ui/tsconfig.json',
          'ui/tsconfig.json',
          'ui/tsconfig.json',
          'ui/tsconfig.json',
          'ui/tsconfig.json',
          'ui/tsconfig.json',
          'ui/tsconfig.node.json',
          'ui/tsconfig.node.json',
          'ui/tsconfig.node.json',
          'ui/tsconfig.node.json',
          'ui/tsconfig.node.json',
          'ui/tsconfig.node.json',
          'ui/tsconfig.node.json',
          'ui/tsconfig.node.json',
          'ui/tsconfig.node.json',
          'ui/tsconfig.node.json',
          'ui/vite.config.ts',
          'ui/vite.config.ts',
          'ui/vite.config.ts',
          'ui/vite.config.ts',
          'ui/vite.config.ts',
          'ui/vite.config.ts',
          'ui/vite.config.ts',
          'ui/vite.config.ts',
          'ui/vite.config.ts',
          'ui/vite.config.ts',
          'ui/yarn.lock',
          'ui/yarn.lock',
          'ui/yarn.lock',
          'ui/yarn.lock',
          'ui/yarn.lock',
          'ui/yarn.lock',
          'ui/yarn.lock',
          'ui/yarn.lock',
          'ui/yarn.lock',
          'ui/yarn.lock']}

install_requires = \
['click>=8.1.3,<9.0.0',
 'fastapi>=0.70.0,<0.71.0',
 'pydantic>=1.9.0,<2.0.0',
 'uvicorn[standard]>=0.15.0,<0.16.0']

entry_points = \
{'console_scripts': ['cado = cado.cli.cli:main']}

setup_kwargs = {
    'name': 'cado',
    'version': '0.1.8',
    'description': 'Python notebook development environment.',
    'long_description': '<div align="center">\n  <img src="assets/cado-banner.png">\n  <h1>cado</h1>\n\n  <p>\n    <strong>Python notebook IDE with a focus on reactivity</strong>\n  </p>\n\n  <br>\n  <div>\n    <a href="https://badge.fury.io/py/cado"><img src="https://badge.fury.io/py/cado.svg" alt="PyPI"></a>\n    <a href="https://pepy.tech/project/cado"><img src="https://pepy.tech/badge/cado" alt="Downloads"></a>\n    <a href="https://github.com/gregorybchris/cado/actions/workflows/ci.yaml"><img src="https://github.com/gregorybchris/cado/actions/workflows/ci.yaml/badge.svg" alt="CI"></a>\n  </div>\n  <br>\n</div>\n\n## About\n\n`cado` is a notebook IDE for Python, like [Jupyter](https://jupyter.org/), but with a reactive cell model, taking inspiration from [Observable](https://observablehq.com/). Each cell defines its own outputs that other cells can listen to. When a child cell runs, it uses cached outputs from parent cells. And when the output of a parent cell updates, the change propagates to all child cells automatically.\n\n## Installation\n\n```bash\npip install cado\n```\n\n## Features\n\nIf you have ideas for new features feel free to create an issue or submit a pull request!\n\n- [x] Reactive cells\n- [x] Auto-save to disk\n- [x] Keyboard shortcuts\n- [x] Drag cells to reorder\n- [x] Markdown mode\n- [x] Notebook files viewer\n\n## Usage\n\n```bash\n# Start up a cado server\ncado up\n\n# Open http://0.0.0.0:8000 in a browser\n```\n\nYou should now see a screen that lists out all of your notebooks. Click on the example notebook to get started.\n\n<p align="center">\n  <img style="inline-block; margin-right: 10px;" src="assets/notebooks-screen.png" height=400>\n  <img style="inline-block" src="assets/notebook-screen.png" height=400>\n</p>\n\n## Keyboard Shortcuts\n\n| Action                                       | Command        |\n| -------------------------------------------- | -------------- |\n| Make the cell above the active cell          | `UpArrow`      |\n| Make the cell below the active cell          | `DownArrow`    |\n| Run the active cell                          | `Shift+Enter`  |\n| Clear the active cell                        | `Shift+Delete` |\n| Turn on edit mode                            | `Enter`        |\n| Turn off edit mode                           | `Escape`       |\n| Create a new cell before the active cell     | `Control+a`    |\n| Create a new cell after the active cell      | `Control+b`    |\n| Create a new cell at the end of the notebook | `Control+n`    |\n| Delete the active cell                       | `Control+d`    |\n',
    'author': 'Chris Gregory',
    'author_email': 'christopher.b.gregory@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/gregorybchris/cado',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8.1,<4',
}


setup(**setup_kwargs)



from datetime import datetime
from pathlib import Path
from uuid import UUID

from pydantic import BaseModel

from cado.core.notebook import Notebook


class NotebookDetails(BaseModel):
    id: UUID
    name: str
    filepath: Path
    created: datetime
    updated: datetime

    @classmethod
    def from_filepath(cls, filepath: Path) -> "NotebookDetails":
        """Load notebook details from a .cado notebook file.

        Args:
            filepath (Path): Filepath to a .cado notebook file.

        Returns:
            NotebookDetails: Loaded notebook details.
        """
        notebook = Notebook.from_filepath(filepath)
        return NotebookDetails(
            id=notebook.id,
            name=notebook.name,
            filepath=filepath,
            created=notebook.created,
            updated=notebook.updated,
        )



from pathlib import Path

from cado.core.notebook import Notebook


def load_example_notebook(filename: str) -> Notebook:
    """Load a simple example notebook.

    Args:
        filename (str): Name of notebook file to load.

    Returns:
        Notebook: An example notebook.
    """
    examples_dirpath = Path(__file__).parent.parent / "examples"
    notebook_filepath = examples_dirpath / filename
    return Notebook.from_filepath(notebook_filepath)
